/*

leases2zone.c

A small dumb program that reads your dhcpd.leases
file and barf out a bunch of A records that you
can put in your named zone file.

Usage: Under the header definitions down there
there's a #define for leases_file, change that
to the location on your system. This was developed
on an OpenBSD machine, I don't know where other
system keep their leases file.

This program comes with absolutely no warranty
implied or explicit. Other than that, do whatever
you want with it. fraggle's arraylist.c/h is part
of his c algorithms library, licensed under the ISC
license.

I'd choose some license for this code but i'm lazy
so .. have fun i guess

version 0.1, jan 31/2011: initial relase

(c) dave kiddell 2011

*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>

#include "arraylist.h"

#define leases_file "/var/db/dhcpd.leases"

#define BUFSIZE 256
unsigned char inbuf[BUFSIZE];
int buflen=0;
unsigned char nextchar='\0';
int filepos=0;

typedef struct
{
	int numeric_addr[4];
	unsigned char hostname[BUFSIZE];
} host_t;

ArrayList *hosts;

host_t cur_host;

enum
{
	PA_TOP=0,
	PA_COMMENT,
	PA_LEASE,
	PA_LEASE_HOSTNAME,
	PA_LEASE_UNRECOGNIZED,
};

int is_singleton(unsigned char whatever)
{
	switch(whatever)
	{
	case '\"':
	case '[':
	case ']':
	case '{':
	case '}':
	case '=':
	case ',':
	case '#':
	case ';':
		return 1;
	}
	return 0;
}

int gettoken(FILE *infile)
{
	buflen = 0;
	if(nextchar == '\0')
	{
		nextchar = fgetc(infile);
		filepos++;
	}

	while(1)
	{
		if(feof(infile))
			return 0;
		if(isspace(nextchar))
		{
			inbuf[buflen] = '\0';
			nextchar = '\0';
			return 1;
		}
		if(is_singleton(nextchar) && buflen==0)
		{
			inbuf[buflen++] = nextchar;
			inbuf[buflen] = '\0';
			nextchar = '\0';
			return 1;
		}

		inbuf[buflen++]=nextchar;
		nextchar = fgetc(infile);
		filepos++;

		if(is_singleton(nextchar))
		{
			inbuf[buflen] = '\0';
			return 1;
		}
	}

	inbuf[buflen] = '\0';
	return 1;
}

int get_nonempty_token(FILE *infile)
{
	buflen=0;
	while(buflen==0)
	{
		if(!gettoken(infile))
			return 0;
	}
	//printf("returning token %s\n",inbuf);
	return 1;
}

int get_expected_char(FILE *infile, unsigned char c, int report_error)
{
	if(!get_nonempty_token(infile))
	{
		printf("Error: Unexpected EOF\n");
		return 0;
	}

	if(inbuf[0] != c)
	{
        if(report_error)
            printf("Error: Expected '%c' but read '%s' (%i) filepos %i\n",c,inbuf,inbuf[0],filepos);
		return 0;
	}
	return 1;
}

int get_expected_singleton(FILE *infile, unsigned char c)
{
    return get_expected_char(infile, c, 1);
}

/*int check_expected_singleton(unsigned char c, int report_error)
{
	if(buflen > 1)
	{
        if(report_error)
            printf("Error: Expected '%c' but read '%s' filepos %i\n",c,inbuf,filepos);
		return 0;
	}
	if(inbuf[0] != c)
	{
        if(report_error)
            printf("Error: Expected '%c' but read '%c' (%i) filepos %i\n",c,inbuf[0],inbuf[0],filepos);
		return 0;
	}
	return 1;
}*/

/*void strip_front_slash(unsigned char *str)
{
    int i;
    for(i=0; i<strlen(str); i++)
    {
        if(str[i] == '/')
            str[i] = '_';
    }
}*/

/*int get_section_id(FILE *infile, unsigned char *sectionname)
{
    int numeric_id;

    // we should get ["something"]={},
    if(!get_nonempty_token(infile))
        return 1;
    // if we got a comma ignore it
    if(inbuf[0] == ',')
        if(!get_nonempty_token(infile))
            return 1;
    // if we got a close brace GTFO
    if(check_expected_singleton('}', 0)) return 1;
    // we must've got an open square bracket then
    if(!check_expected_singleton('[', 1)) return 1;
    //if(!get_expected_singleton(infile,'\"')) return 1;
    // if it not in quotes then its numeric~
    if(!get_expected_char(infile,'\"',0))
    {
        numeric_id=1;
    }else{
        numeric_id=0;
        if(!get_nonempty_token(infile))
            return 1;
    }
    strlcpy(sectionname,inbuf,BUFZIE);
    if(!numeric_id)
        if(!get_expected_singleton(infile,'\"'))
            return 1;

    if(!get_expected_singleton(infile,']')) return 1;
    if(!get_expected_singleton(infile,'=')) return 1;

    return 0;
}*/

int compare_hosts(ArrayListValue a, ArrayListValue b)
{
	host_t *host1 = (host_t *)a;
	host_t *host2 = (host_t *)b;
	return host1->numeric_addr[3] - host2->numeric_addr[3];
}

int identical_hosts(ArrayListValue a, ArrayListValue b)
{
	host_t *host1 = (host_t *)a;
	host_t *host2 = (host_t *)b;
	if(host1->numeric_addr[3] == host2->numeric_addr[3] &&
		!strcmp(host1->hostname, host2->hostname))
	{
		return 1;
	}
	return 0;
}

void parse_file(int state, FILE *infile)
{
	time_t mytime;
	struct tm *loctime;
	int i;
	FILE *outfile;
	host_t *host_to_store;

	switch(state)
	{
	case PA_TOP:
		// we will either see comments or the
		// start of a lease
		if(!get_nonempty_token(infile))
			return;
		
		// is it a comment?
		if(inbuf[0] == '#')
		{
			parse_file(PA_COMMENT, infile);
		}else if(!strcmp(inbuf,"lease")){
			if(!get_nonempty_token(infile))
			{
				printf("error parsing. bailing.\n");
				return;
			}
			sscanf(inbuf, "%d.%d.%d.%d",
						&(cur_host.numeric_addr[0]),
						&(cur_host.numeric_addr[1]),
						&(cur_host.numeric_addr[2]),
						&(cur_host.numeric_addr[3]));
			cur_host.hostname[0] = '\0';
			/*printf("%d.%d.%d.%d\n",
						cur_host.numeric_addr[0],
						cur_host.numeric_addr[1],
						cur_host.numeric_addr[2],
						cur_host.numeric_addr[3]);*/

			if(!get_expected_singleton(infile,'{')) return;
			parse_file(PA_LEASE, infile);
		}
		
        parse_file(PA_TOP, infile);
		break;

	case PA_COMMENT:
		// discard everything until we see newline
		if(nextchar == '\n')
			return;
		do {
			nextchar = fgetc(infile);
			filepos++;
		} while (nextchar != '\n');
		return;
	case PA_LEASE:
		
		if(!get_nonempty_token(infile))
			return;
		
		// if the field is the hostname then we are
		// interested
		if(!strcmp(inbuf, "client-hostname"))
		{
			parse_file(PA_LEASE_HOSTNAME, infile);
			parse_file(PA_LEASE, infile);
			return;
		}
		
		// if the field is close brace then we are done
		if(inbuf[0] == '}')
			return;
		
		// if we get here then we don't know what field we
		// just read so we will send the parser looking
		// for a semicolon using PA_LEASE_UNRECOGNIZED state
		parse_file(PA_LEASE_UNRECOGNIZED, infile);
		parse_file(PA_LEASE, infile);
		break;
	
	case PA_LEASE_HOSTNAME:
		if(!get_expected_singleton(infile,'\"')) return;
		if(!get_nonempty_token(infile)) return;

		//printf("hostname for this lease is %s\n", inbuf);
	
		host_to_store = malloc(sizeof(host_t));
		strlcpy(host_to_store->hostname, inbuf, BUFSIZE);
		host_to_store->numeric_addr[0] = cur_host.numeric_addr[0];
		host_to_store->numeric_addr[1] = cur_host.numeric_addr[1];
		host_to_store->numeric_addr[2] = cur_host.numeric_addr[2];
		host_to_store->numeric_addr[3] = cur_host.numeric_addr[3];
	
		// check to make sure we don't already have this in the hosts arraylist
		i = arraylist_index_of(hosts, &identical_hosts, host_to_store);
		//printf("index of %s in arraylist is %i\n", host_to_store->hostname, i);
		if(i == -1)
		{
			arraylist_append(hosts, host_to_store);
		}else{
			free(host_to_store);
		}
	
		if(!get_expected_singleton(infile,'\"')) return;
		if(!get_expected_singleton(infile,';')) return;
		break;
	case PA_LEASE_UNRECOGNIZED:
		// discard everything until we see semicolon
		if(inbuf[0] == ';')
			return;
		do {
			if(!get_nonempty_token(infile))
				return;
		} while (inbuf[0] != ';');
		break;
	default:
		printf("Error: unrecognized parsing state\n");
		return;
	}
}

void reset_global_vars()
{
    filepos=0;
    buflen=0;
    inbuf[0]='\0';
    nextchar='\0';
}

int main(int argc, char *argv[])
{
	FILE *infile;
	int i;
	host_t *host_iter;

	hosts = arraylist_new(0);
    reset_global_vars();
	infile = fopen(leases_file, "rb");
    parse_file(PA_TOP, infile);
	fclose(infile);

	arraylist_sort(hosts, &compare_hosts);
	
	for(i=0; i<hosts->length; i++)
	{
		host_iter = (host_t*)hosts->data[i];
		printf("%s", host_iter->hostname);
		if(strlen(host_iter->hostname) < 8)
			printf("\t\t\t");
		else
			printf("\t\t");
		
		printf("A\t%d.%d.%d.%d\n", 
			 host_iter->numeric_addr[0],
			 host_iter->numeric_addr[1],
			 host_iter->numeric_addr[2],
			 host_iter->numeric_addr[3]
		);
	}
	

	return 0;
}
